import { mat4, vec2, vec3, vec4 } from "gl-matrix";
import * as util from "./gl_util";

export class VertexBuffer{
    readonly id : WebGLBuffer;

    constructor(id : WebGLBuffer){
        this.id = id;
    }
    bind(gl:WebGL2RenderingContext){
        gl.bindBuffer(gl.ARRAY_BUFFER,this.id);
    }
    unbind(gl:WebGL2RenderingContext){
        gl.bindBuffer(gl.ARRAY_BUFFER,null);
    }
}

/***
 * create vertex buffer
 * @param gl GLContext
 * @param data vertex data to draw
 * @example ```
 * const buffer = createVertexBuffer(gl,[
 * -0.5,0.5,
 * 0.5,0.5,
 * 0.5,-0.5,
 * -0.5,-0.5]);
 * ```
 */
export function createVertexBuffer(gl:WebGL2RenderingContext, data:number[]): VertexBuffer{
    const id = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER,id);
    gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(data),gl.STATIC_DRAW);
    return new VertexBuffer(id);
}

export class IndexBuffer{
    readonly id: WebGLBuffer;
    readonly count:number;
    constructor(id: WebGLBuffer, count:number){
        this.id = id;
        this.count = count;
    }
    bind(gl:WebGL2RenderingContext){
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER,this.id);
    }
    unbind(gl:WebGL2RenderingContext){
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER,null);
    }
}

export function createIndexBuffer(gl:WebGL2RenderingContext, data:number[]){
    const id = gl.createBuffer();
    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER,id);
    gl.bufferData(gl.ELEMENT_ARRAY_BUFFER,new Uint16Array(data),gl.STATIC_DRAW);
    return new IndexBuffer(id,data.length);
}

export class GLProgram{
    readonly program: WebGLProgram;
    #uniformLoc : Map<string,WebGLUniformLocation>;
    constructor(program: WebGLProgram){
        this.program = program;
        this.#uniformLoc = new Map<string,WebGLUniformLocation>();
    }
    getActiveUniforms(gl:WebGL2RenderingContext):WebGLActiveInfo[]{
        const num = gl.getProgramParameter(this.program,gl.ACTIVE_UNIFORMS);
        const info = [...new Array(num).keys()].map(i=>gl.getActiveUniform(this.program,i));
        return info;
    }
    getActiveAttributes(gl:WebGL2RenderingContext):WebGLActiveInfo[]{
        const num = gl.getProgramParameter(this.program,gl.ACTIVE_ATTRIBUTES);
        const info = [...new Array(num).keys()].map(i=>gl.getActiveAttrib(this.program,i));
        return info;
    }
    getAttribLocation(gl:WebGL2RenderingContext,name:string):number{
        return gl.getAttribLocation(this.program,name);
    }
    getUniformLocation(gl:WebGL2RenderingContext,name:string):WebGLUniformLocation{
        if(this.#uniformLoc.has(name)){
            return this.#uniformLoc.get(name);
        }
        else{
            const location = gl.getUniformLocation(this.program,name);
            this.#uniformLoc.set(name,location);
            return location;
        }
    }
    use(gl:WebGL2RenderingContext):void{
        gl.useProgram(this.program);
    }
    
    setUniform1i(gl:WebGL2RenderingContext,name:string,v:number){
        const loc = this.getUniformLocation(gl,name);
        gl.uniform1i(loc,v);
    }
    setUniform1f(gl:WebGL2RenderingContext,name:string,v:number){
        const loc = this.getUniformLocation(gl,name);
        gl.uniform1f(loc,v);
    }
    setUniform2fv(gl:WebGL2RenderingContext,name:string,v:vec2){
        const loc = this.getUniformLocation(gl,name);
        gl.uniform2fv(loc,v);
    }
    setUniform3fv(gl:WebGL2RenderingContext,name:string,v:vec3){
        const loc = this.getUniformLocation(gl,name);
        gl.uniform3fv(loc,v);
    }
    setUniform4fv(gl:WebGL2RenderingContext,name:string,v:vec4){
        const loc = this.getUniformLocation(gl,name);
        gl.uniform4fv(loc,v);
    }
    setUniformMat4f(gl:WebGL2RenderingContext,name:string,v:mat4){
        const loc = this.getUniformLocation(gl,name);
        gl.uniformMatrix4fv(loc,false,v);
    }
    
    unuse(gl:WebGL2RenderingContext):void{
        gl.useProgram(null);
    }
}

export class VertexArray{
    readonly vao:WebGLVertexArrayObject;
    constructor(vao:WebGLVertexArrayObject){
        this.vao = vao;
    }
    bind(gl:WebGL2RenderingContext){
        gl.bindVertexArray(this.vao);
    }
    unbind(gl:WebGL2RenderingContext){
        gl.bindVertexArray(null);
    }
    addBuffer(gl:WebGL2RenderingContext,va:VertexBuffer,loc:number,layout:{
        /**count of one element */
        count: GLint,
        type: GLenum,
        normalized: GLboolean,
        stride: GLsizei,
        offset: GLintptr
    }){
        this.bind(gl);
        va.bind(gl);
        gl.enableVertexAttribArray(loc);
        gl.vertexAttribPointer(loc,layout.count,layout.type,layout.normalized,layout.stride,layout.offset);
    }
}
export function createVertexArray(gl:WebGL2RenderingContext){
    return new VertexArray(gl.createVertexArray());
}